
from edge import eventlog
from edge.fov import Fov
from edge.service import react, gpsd, tz, about, db_wanted
from edge.lane_detector import LaneDetector

from _kernel_args import KernelArgs
from .model.models import Track

from .services import (stats, event, trainer, raas, bs, plate_info, vs,
                       tl, leddar, db, systest, clock, sign)

from .detectors.types import IBaseDetector, IDetector
from .detectors.base.detector import create as mk_base_detector
from .detectors.transit.detector import create as mk_transit_detector
from .detectors.parking.detector import create as mk_parking_detector
from .detectors.wrong_way.detector import create as mk_wrong_way_detector
from .detectors.spot_speed.detector import create as mk_spot_speed_detector
from .detectors.red_light.detector import create as mk_red_ligth_detector
from .detectors.red_light2.detector import create as mk_red_ligth_detector2
from .detectors.section_speed.detector import create as mk_section_speed_detector
from .detectors.wanted.detector import create as mk_wanted_detector

from .model.types import BaseEvent, Event


class BaseKernel:
    def __init__(self, stat_service, react_services, event_service, leddar_service, base_detector, detectors_detected, detectors_lost,
                 detectors_no_plate):
        """
        :type stat_service: stats.IStatService
        :type react_services: list[react.IReactService]
        :type event_service: event.IEventService
        :type leddar_service: leddar.ILeddarService
        :type base_detector: IBaseDetector
        :type detectors_detected: list[IDetector]
        :type detectors_lost: list[IDetector]
        :type detectors_no_plate: list[IDetector]
        """
        self._event_service = event_service
        self._leddar_service = leddar_service

        self._detectors_detected = detectors_detected
        self._detectors_lost = detectors_lost
        self._detectors_no_plate = detectors_no_plate
        map(lambda x: x.set_handler(self._handle_event), detectors_no_plate)
        map(lambda x: x.set_handler(self._handle_event), detectors_detected)
        map(lambda x: x.set_handler(self._handle_event), detectors_lost)

        self._base_detector = base_detector
        self._base_detector.set_handler(self._handle_base_event)

        self._react_services = react_services
        for react_service in self._react_services:
            react_service.set_handler(self._handle_react)

        self._react_counters = stat_service.react

    def _handle_react(self, react_name, react_jnode):
        """
        :type react_name: str
        :type react_jnode: dict
        """
        try:
            tracks = Track.deserialize(react_jnode)
        except (KeyError, ValueError) as x:
            eventlog.log_e('fail: unpack {} <{}>'.format(react_name, x.message))
            return

        for track in tracks:
            if track.le_track and len(track.le_track):
                self._leddar_service.update(react_jnode)

            self._react_counters.inc_track(track)

            lr_track = track.lr_track
            if lr_track:
                init_quad = lr_track[0].plate_quad
                fini_quad = lr_track[-1].plate_quad
                self._react_counters.inc_plate_quad(init_quad, fini_quad)

            self._base_detector.analyze(track)

    def _handle_base_event(self, track, event_args):
        """
        :type track: Track
        :type event_args: BaseEvent
        """
        if track.lr_track is None:
            map(lambda x: x.analyze(track, event_args), self._detectors_no_plate)

        elif track.lr_track.stat_detected:
            map(lambda x: x.analyze(track, event_args), self._detectors_detected)

        elif track.lr_track.stat_lost:
            map(lambda x: x.analyze(track, event_args), self._detectors_lost)

    def _handle_event(self, event_args):
        """
        :type event_args: Event
        """
        self._event_service.emit(event_args)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        sign_service = sign.init_service(args.sign)

        stat_service = stats.init_service(
            twisted_reactor, args.counters, [sign_service],
            stat_gpsd=bool(args.gpsd.endpoint))

        db_service = db.init_service(args.db)
        db_wanted_service = db_wanted.init_service(args.db.db_args)

        systest_service = systest.init_service(args.systest)

        lane_service = LaneDetector(
            args.lane_detector.id_to_lane_list,
            args.lane_detector.img_width,
            args.lane_detector.img_height)

        fov_service = Fov.create2(args.image_mapper) if args.image_mapper else None

        tz_service = tz.init_service(args.tz)
        bs_service = bs.init_service(twisted_reactor, args.bs, stat_service.bs)
        vs_service = vs.init_service(args.vs)
        clock_service = clock.init_service(twisted_reactor)

        raas_service = raas.init_service(twisted_reactor, args.raas)
        tl_service = tl.init_service(twisted_reactor, args.tl)
        leddar_service = leddar.init_service(args.ledar)  # TODO
        trainer_service = trainer.init_service(twisted_reactor, args.trainer, tz_service)
        gpsd_service = gpsd.init_service(twisted_reactor, args.gpsd, stat_service.gpsd)
        about_service = about.init_service(args.about, gpsd_service)
        event_service = event.init_service(
            twisted_reactor, stat_service.event, args.event_service, tz_service, bs_service,
            systest_service, about_service, sign_service)

        plate_info_service = plate_info.init_service(args.plate_info)

        base_detector = mk_base_detector(
            tz_service, vs_service, raas_service, trainer_service,
            about_service, lane_service, plate_info_service, None, clock_service, args.base_detector)  # TODO: check fov service arg

        detectors_detected = list()
        detectors_no_plate = list()
        detectors = list()
        if not args.transit_args.require_plate:
            detectors_no_plate.append(mk_transit_detector(about_service, tz_service, clock_service, args.transit_args))
            detectors.append(mk_transit_detector(about_service, tz_service, clock_service, args.transit_args))
        elif args.transit_args.plate_detected_mode:
            detectors_detected.append(mk_transit_detector(about_service, tz_service, clock_service, args.transit_args))
        else:
            detectors.append(mk_transit_detector(about_service, tz_service, clock_service, args.transit_args))

        detectors.extend(map(
            lambda p_args: mk_parking_detector(tz_service, vs_service, about_service, db_service, lane_service, p_args),
            args.detectors.parking))

        detectors.extend(map(
            lambda ww_args: mk_wrong_way_detector(
                twisted_reactor, tz_service, vs_service, about_service, lane_service, ww_args),
            args.detectors.wrong_way))

        detectors.extend(map(
            lambda ss_args: mk_spot_speed_detector(
                twisted_reactor, tz_service, vs_service, about_service, lane_service, ss_args),
            args.detectors.spot_speed))

        detectors.extend(map(
            lambda rl_args: mk_red_ligth_detector(
                twisted_reactor, tz_service, vs_service, about_service, tl_service, lane_service, rl_args),
            args.detectors.red_light))

        detectors_no_plate.extend(map(
            lambda rl_args: mk_red_ligth_detector2(
                twisted_reactor, tz_service, vs_service, about_service, tl_service, leddar_service, lane_service, rl_args),
            args.detectors.railroad_crossing))

        detectors.extend(map(
            lambda wanted_args: mk_wanted_detector(about_service, db_wanted_service, wanted_args),
            args.detectors.wanted))

        detectors.extend(map(
            lambda ss_args: mk_section_speed_detector(tz_service, vs_service, about_service, clock_service, fov_service, ss_args),
            args.detectors.section_speed))

        react_services = [
            react.init_service(twisted_reactor, [args.react_service.react_endpoint])
        ]
        if args.ledar:
            react_services.append(react.init_service(twisted_reactor, [args.ledar.endpoint]))

        def shutdown():
            for react_service in react_services:
                react.fini_service(react_service)

            map(lambda detector: detector.destroy(), detectors)
            map(lambda detector: detector.destroy(), detectors_no_plate)
            map(lambda detector: detector.destroy(), detectors_detected)
            base_detector.destroy()

            event.fini_service(event_service)
            about.fini_service(about_service)
            gpsd.fini_service(gpsd_service)
            trainer.fini_service(trainer_service)
            raas.fini_service(raas_service)
            tl.fini_service(tl_service)
            leddar.fini_service(leddar_service)
            bs.fini_service(bs_service)
            tz.fini_service(tz_service)
            vs.fini_service(vs_service)
            db.fini_service(db_service)
            db_wanted.fini_service(db_wanted_service)
            stats.fini_service(stat_service)
            sign.fini_service(sign_service)
            systest.fini_service(systest_service)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

        BaseKernel.__init__(self, stat_service, react_services, event_service, leddar_service, base_detector, detectors_detected, detectors, detectors_no_plate)
        self._twisted_reactor = twisted_reactor
        self._counters = stats

    def run(self):
        self._twisted_reactor.run()
