from edge_bbox.types import RectI, QuadI
from edge.service.types import Geolocation
from edge.service.tz import ITzService
from edge.model.types import Dir, Origin, Location
from edge.image.types import MemBlob, Blob

from ..services.plate_info import IPlateInfoService

from uuid import uuid4


class Identity:
    def __init__(self, sender_type, sender_id):
        """
        :type sender_type: int
        :type sender_id: int
        """
        self.type = sender_type
        self.id = sender_id

        self._hash = sender_type << 14 + sender_id

    def __str__(self):
        return '%s.%s' % (self.type, self.id)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return getattr(other, 'type', None) == self.type and \
               getattr(other, 'id', None) == self.id and \
               isinstance(other, self.__class__)


class LicensePlateType:
    def __init__(self, template_ref, plate_info_service=None):
        """
        :type template_ref: int
        :type plate_info_service: IPlateInfoService
        """
        self._numeric = plate_info_service.get_country_code(template_ref)
        self._draft_jnode = {
            'country': {
                'numeric': self._numeric
            }
        }

        plate_attrs = plate_info_service.get_plate_attrs(template_ref)
        if plate_attrs:
            self._draft_jnode['info'] = [{
                'attributes': plate_attrs
            }]

        self._jnode = None

    @property
    def country_numeric(self):
        """
        :rtype: int
        """
        return self._numeric

    def serialize(self):
        """
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class LicensePlate:
    def __init__(self, validity, text, plate_type, quad):
        """
        :type validity: int
        :type text: str
        :type plate_type: LicensePlateType
        :type quad: QuadI
        """
        self.plate_text = text
        self._draft_jnode = {
            'weight': validity,
            'type': plate_type.serialize(),
            'bbox-2': quad.get_abs_bbox(),
            'bbox-4': quad.serialize(),
            'text': {
                'ucode': text
            }

        }
        self._plate_type = plate_type
        self._jnode = None

    @property
    def country_numeric(self):
        """
        :rtype: int
        """
        return self._plate_type.country_numeric

    def serialize(self):
        """
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class Position:
    def __init__(self, forward, left):
        """
        :type forward: float
        :type left: float
        """
        self.forward = forward
        self.left = left

    def serialize(self):
        """
        :rtype: dict
        """
        jnode = {
            'forward': self.forward,
            'left': self.left
        }
        return jnode


class Vehicle:
    def __init__(self, licence_plates, spot_speed, lenght, direction, video_speed2=None, position=None,
                 type_jnode=None, systest=False, lpr_speed=None):
        """
        :type licence_plates: list[LicensePlate]
        :type spot_speed: int | None
        :type lenght: float | None
        :type direction: Dir
        :type video_speed2: int | None
        :type position: Position | None
        :type type_jnode: dict | None
        :type systest: bool
        """
        self.position = position
        self.direction = direction
        self._camera_speed = None
        self._video_speed2 = None
        self._spot_speed = None
        self._lpr_speed = lpr_speed

        self._jnode = None
        self._draft_jnode = {
            'license-plates': map(lambda plate: plate.serialize(), licence_plates)
        }

        self._type_jnode = type_jnode
        if type_jnode is not None:
            self._draft_jnode["type"] = type_jnode

        if direction.dir != Dir.UNKNOWN:
            self._draft_jnode['dir'] = str(direction)

        # if systest video_speed2 is not required
        if spot_speed and (systest or video_speed2):

            if spot_speed > 240:
                ra_speed2_str = '>240'
            elif spot_speed < 20:
                ra_speed2_str = '<20'
            else:
                ra_speed2_str = str(spot_speed)

            if systest:
                self._spot_speed = spot_speed
                self._draft_jnode['spot-speed'] = {
                    'radar': {
                        'measured': spot_speed,
                        'measured2': ra_speed2_str
                    }
                }

            elif 0.9 * spot_speed < abs(video_speed2) < 1.1 * spot_speed:
                from random import randint
                delta_kmh = 8
                delta_p = 0.08
                delta = max(delta_kmh, int(spot_speed * delta_p))
                self._camera_speed = spot_speed + randint(-delta, delta)
                self._video_speed2 = abs(video_speed2)
                self._spot_speed = spot_speed

                # self._draft_jnode['spot-speed'] = {'radar': {'measured': spot_speed}}

                self._draft_jnode['spot-speed'] = {
                    'radar': {
                        'measured': spot_speed,
                        'measured2': ra_speed2_str
                    },
                    'camera': {
                        'measured': self._camera_speed,
                        'measured2': self._video_speed2
                    }
                }

                if self._lpr_speed is not None:
                    self._draft_jnode['spot-speed']['camera']['measured3'] = self._lpr_speed

        elif self._lpr_speed:
            self._draft_jnode['spot-speed'] = {
                'camera': {
                    'measured3': self._lpr_speed
                }
            }

        if lenght:
            self._draft_jnode['size'] = {'l': lenght}

        if position:
            self._draft_jnode['position'] = position.serialize()

        self.plate_absent = len(licence_plates) == 0
        self._plates = licence_plates

    @property
    def vehicle_class(self):
        """
        :rtype: str | None
        """
        if self._type_jnode:
            return self._type_jnode['info'][0]['class']

    @property
    def vehicle_make(self):
        """
        :rtype: str | None
        """
        if self._type_jnode:
            return self._type_jnode['make']

    @property
    def vehicle_model(self):
        """
        :rtype: str | None
        """
        if self._type_jnode:
            return self._type_jnode['model']

    @property
    def plate_text(self):
        """
        :rtype: str | None
        """
        if self._plates:
            return self._plates[0].plate_text

    @property
    def plate_county_numeric(self):
        """
        :rtype: int | None
        """
        if self._plates:
            return self._plates[0].country_numeric

    @property
    def spot_speed(self):
        """
        :rtype: int | None
        """
        return self._spot_speed

    @property
    def camera_speed(self):
        """
        :rtype: int | None
        """
        return self._camera_speed

    @property
    def video_speed2(self):
        """
        :rtype: int | None
        """
        return self._video_speed2

    def serialize(self):
        """
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = self._draft_jnode

        return self._jnode


class Waypoint:
    """
    :type _jnode: dict | None
    """

    NAME_CATCH = 'catch'
    NAME_BEST = 'best'
    NAME_LOST = 'lost'
    NAME_VIOLATION = 'violation'
    NAME_VIOLATION_BEST = 'violation-best'
    NAME_PRE_STOPBAR = 'pre-stopbar'
    NAME_POST_STOPBAR = 'post-stopbar'

    def __init__(self, name, ts, control_ts=None, location=Location(), draft_jnode={}):
        """
        :type name: str
        :type ts: long
        :type control_ts: long | None
        :type location: Location
        """
        self._name = name
        self._ts = ts
        self._control_ts = control_ts
        self._images = []

        self._draft_jnode = draft_jnode
        location.serialize(self._draft_jnode)
        self._jnode = None

    @property
    def name(self):
        return self._name

    @property
    def ts(self):
        return self._ts

    def add_images(self, blobs):
        """
        :type blobs: list[MemBlob]
        :rtype: Waypoint
        """
        self._images.extend(blobs)
        return self

    def get_blobs(self):
        """
        :rtype: list of MemBlob
        """
        return self._images

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = {'time': tz_service.get_datetime_isostr(self._ts / 1e6)}
            if self._control_ts is not None:
                self._jnode['control-time'] = tz_service.get_datetime_isostr(self._control_ts / 1e6)

            self._jnode.update(self._draft_jnode)

        images_jnode = dict()
        for blob in self._images:  # type: MemBlob
            images_jnode[blob.profile] = blob.serialize(tz_service=tz_service)

        if len(images_jnode):
            self._jnode['images'] = images_jnode

        return self._jnode


class BestWaypoint(Waypoint):
    """
    :type _images: list of MemBlob
    :type _jnode: dict | None
    """

    DIR_UNKNOWN = 'unknown'
    DIR_APPROACHING = 'approching'
    DIR_RECEDING = 'receding'

    def __init__(self, ts, vehicle, control_ts=None, direction='unknown', location=Location()):
        """
        :type ts: long
        :type vehicle: Vehicle
        :type control_ts: long | None
        :type direction: str
        :type location: Location
        """
        Waypoint.__init__(self, Waypoint.NAME_BEST, ts, control_ts=control_ts, location=location)

        self._images = []
        self._draft_jnode = {
            'vehicle': vehicle.serialize()
        }

        if direction != 'unknown':
            self._draft_jnode['dir'] = direction

        self.vehicle = vehicle
        self.location = location

        self._jnode = None

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

            self.location.serialize(self._jnode)

        images_jnode = dict()
        images_jnode['license-plates'] = []
        for blob in self._images:
            if blob.profile == 'plate':
                images_jnode['license-plates'].append(blob.serialize(tz_service=tz_service))
            else:
                images_jnode[blob.profile] = blob.serialize(tz_service=tz_service)

        if len(images_jnode):
            self._jnode['images'] = images_jnode

        return self._jnode


class ViolationWaypoint(Waypoint):
    def __init__(self, ts, location, control_ts=None, position=None, direction=None, plate_rects=list(), plate_quads=list(),
                 w_name=Waypoint.NAME_VIOLATION):
        """
        :type ts: long
        :type location: Location
        :type control_ts: long | None
        :type position: Position | None
        :type direction: Dir | None
        :type plate_rects: list[RectI]
        :type plate_quads: list[QuadI]
        :type w_name: str
        :rtype: ViolationWaypoint
        """
        Waypoint.__init__(self, w_name, ts, control_ts=control_ts, location=location)
        self._draft_jnode = {
          'vehicle': {
              'position': position.serialize()
          }
        } if position else {}

        if position or len(plate_rects) or len(plate_quads):
            self._draft_jnode = {
                'vehicle': {}
            }

            if direction and direction.dir != Dir.UNKNOWN:
                self._draft_jnode['vehicle']['dir'] = str(direction)

            if position:
                self._draft_jnode['vehicle']['position'] = position.serialize()

            if len(plate_quads):
                self._draft_jnode['vehicle']['licence-plates'] = map(
                    lambda quad: {'bbox4': quad.serialize(),
                                  'bbox2': quad.get_abs_bbox()}, plate_quads)
            elif len(plate_rects):
                self._draft_jnode['vehicle']['licence-plates'] = map(
                    lambda rect: {'bbox2': rect.serialize()}, plate_rects)

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

        return self._jnode


class BestViolationWaypoint(ViolationWaypoint):
    def __init__(self, ts, location, control_ts=None, position=None, direction=None, plate_rects=list(), plate_quads=list()):
        """
        :type ts: long
        :type location: Location
        :type control_ts: long | None
        :type position: Position | None
        :type direction: Dir | None
        :type plate_rects: list[RectI]
        :type plate_quads: list[QuadI]
        :rtype: BestViolationWaypoint
        """
        ViolationWaypoint.__init__(self, ts, location, control_ts=control_ts, position=position, direction=direction,
                                   plate_rects=plate_rects, plate_quads=plate_quads, w_name=Waypoint.NAME_VIOLATION_BEST)


class RedLightViolationWaypoint(Waypoint):
    def __init__(self, name, ts, location, plate_quads, green_amber_time, red_time, red_amber_time):
        """
        :type name: str
        :type ts: long
        :type location: Location
        :type plate_quads: list[QuadI]
        :param green_amber_time: value in usec
        :type green_amber_time: int | None

        :param red_time: value in usec
        :type red_time: int | None

        :param red_amber_time: value in usec
        :type red_amber_time: int | None
        """
        Waypoint.__init__(self, name, ts, location=location)
        self._draft_jnode = {
            'vehicle': {
                'licence-plates': map(lambda quad: {
                    'bbox2': quad.get_abs_bbox(),
                    'bbox4': quad.serialize()
                }, plate_quads)
            }
        }

        self._green_amber_time = self._red_time = self._red_amber_time = 0
        if green_amber_time is not None:
            self._draft_jnode['green-amber-time'] = green_amber_time / 1000
            self._green_amber_time = green_amber_time / 1000

        if red_time is not None:
            self._draft_jnode['red-time'] = red_time / 1000
            self._red_time = red_time / 1000

        if red_amber_time is not None:
            self._draft_jnode['red-amber-time'] = red_amber_time / 1000
            self._red_amber_time = red_amber_time / 1000

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if self._jnode is None:
            self._jnode = Waypoint.serialize(self, tz_service)
            self._jnode.update(self._draft_jnode)

        return self._jnode

    @property
    def red_time(self):
        """
        :rtype: int
        """
        return self._red_time

    @property
    def green_amber_time(self):
        """
        :rtype: int
        """
        return self._green_amber_time

    @property
    def red_amber_time(self):
        """
        :rtype: int
        """
        return self._red_amber_time


class Measurement:
    """
    :type _waypoints: list of Waypoint
    :type _jnode: dict
    """
    def __init__(self, uuid, origin=None, direction=None):
        """
        :type uuid: str
        :type origin: Origin | None
        :type direction: Dir | None
        """
        self._waypoints = []
        self._jnode = {
            'uuid': uuid
        }

        if origin:
            self._jnode['origin'] = origin.serialize()

        if direction:
            self._jnode['vehicle'] = {'dir': str(direction)}

    @property
    def best_waypoint(self):
        """
        :rtype: BestWaypoint | None
        """
        bw = filter(lambda point: point.name == point.NAME_BEST, self._waypoints)
        return bw[0] if len(bw) == 1 else None

    @property
    def waypoints(self):
        """
        :rtype: list[Waypoint]
        """
        return self._waypoints

    def add_waypoints(self, waypoints):
        """
        :type waypoints: list[Waypoint]
        :rtype: Measurement
        """
        self._waypoints.extend(waypoints)
        return self

    def get_blobs(self):
        """
        :rtype: list of MemBlob
        """
        blobs = []
        map(lambda waypoint: blobs.extend(waypoint.get_blobs()), self._waypoints)

        return blobs

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        waypoints_jnode = dict()
        for waypoint in self._waypoints:
            waypoints_jnode[waypoint.name] = waypoint.serialize(tz_service)

        self._jnode['waypoints'] = waypoints_jnode

        return self._jnode


class BaseEvent:
    """
    :type _measurements: list[Measurement]
    """
    def __init__(self, measurements, alarms_jnode):
        """
        :type measurements: list[Measurement]
        :type alarms_jnode: list[dict]| None
        """
        self._measurements = list(measurements)
        self.alarms_jnode = alarms_jnode

    @property
    def measurements(self):
        """
        :rtype: list[Measurement]
        """
        return self._measurements

    @property
    def best_wp(self):
        """
        :rtype: BestWaypoint | None
        """
        waypoints_list = self._measurements[0].waypoints
        best_points = filter(
            lambda waypoint: waypoint.name == waypoint.NAME_BEST,
            waypoints_list)
        if len(best_points) != 1:
            return None

        return best_points[0]

    def get_blobs(self):
        """
        :rtype: list[MemBlob]
        """
        blobs = []
        map(lambda measurement: blobs.extend(measurement.get_blobs()), self._measurements)

        return blobs

    def get_plate_text(self):
        """
        :rtype: str | None
        """
        waypoints = self.measurements[0].waypoints
        best_waypoint_list = filter(lambda waypoint: waypoint.name == waypoint.NAME_BEST, waypoints)

        best_waypoint = best_waypoint_list[0]  # type: BestWaypoint
        return best_waypoint.vehicle.plate_text

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if len(self._measurements) != 1:
            raise NotImplementedError

        jnode = {'measurements': [self._measurements[0].serialize(tz_service)]}
        if self.alarms_jnode:
            jnode['alarms'] = self.alarms_jnode

        return jnode


class Evidence:
    """
    :type _jnode: dict | None
    :type _waypoints: list[Waypoint]
    :type _videos: list[FileBlob]
    """

    NAME_TRANSIT = 'transit'
    NAME_PARKING = 'parking'
    NAME_RED_LIGHT = 'red-light'
    NAME_WRONG_WAY = 'wrong-way'
    NAME_SPOT_SPEED = 'spot-speed'
    NAME_WANTED = 'wanted'
    NAME_SECTION_SPEED = 'section-speed'

    def __init__(self, name, origin, ts, control_ts=None):
        """
        :type name: str
        :type origin: Origin
        :type ts: long
        :type control_ts: long
        """
        self._ts = ts
        self._control_ts = control_ts

        self._draft_jnode = {
            'type': name,
            'origin': origin.serialize()
        }
        self._origin = origin
        self._type = name

        self._waypoints = []
        self._jnode = None
        self._videos = []

    @property
    def origin(self):
        """
        :rtype: Origin
        """
        return self._origin

    @property
    def type(self):
        """
        :rtype: str
        """
        return self._type

    @property
    def ts(self):
        return self._ts

    def add_waypoints(self, waypoints):
        """
        :type waypoints: list[Waypoint]
        """
        self._waypoints.extend(waypoints)

    def add_videos(self, videos):
        """
        :type videos: list[Blob]
        """
        self._videos.extend(videos)

    @property
    def videos(self):
        """
        :rtype list[Blob]
        """
        return self._videos

    def get_blobs(self):
        """
        :rtype: list[Blob]
        """
        blobs = []
        blobs.extend(self._videos)

        for waypoint in self._waypoints:
            blobs.extend(waypoint.get_blobs())
        return blobs

    @property
    def waypoints(self):
        """
        :rtype: list[Waypoint]
        """
        return self._waypoints

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        if self._jnode is None:
            self._draft_jnode['time'] = tz_service.get_datetime_isostr(self._ts / 1e6)
            if self._control_ts is not None:
                self._draft_jnode['control-time'] = tz_service.get_datetime_isostr(self._control_ts / 1e6)

            self._jnode = self._draft_jnode

        return self._jnode


class Event:
    TYPE_TRANSIT = Evidence.NAME_TRANSIT
    TYPE_PARKING = Evidence.NAME_PARKING
    TYPE_RED_LIGHT = Evidence.NAME_RED_LIGHT
    TYPE_WRONG_WAY = Evidence.NAME_WRONG_WAY
    TYPE_SPOT_SPEED = Evidence.NAME_SPOT_SPEED
    TYPE_WANTED = Evidence.NAME_WANTED
    TYPE_SECTION_SPEED = Evidence.NAME_SECTION_SPEED

    def __init__(self, sender, base, evidence):
        """
        :type sender: Identity
        :type base: BaseEvent
        :type evidence: Evidence
        """
        self._uuid = str(uuid4())
        self._sender = sender
        self._base = base
        self._evidence = evidence

    @property
    def uuid(self):
        """
        :rtype: str
        """
        return self._uuid

    @property
    def base(self):
        """
        :rtype: BaseEvent
        """
        return self._base

    @property
    def evidence(self):
        """
        :rtype: Evidence
        """
        return self._evidence

    @property
    def type(self):
        """
        :rtype: str
        """
        return self._evidence.type

    @property
    def sender(self):
        return self._sender

    @property
    def ts_usec(self):
        return self._evidence.ts / 1000

    @property
    def ts(self):
        return self._evidence.ts

    def get_blobs(self):
        """
        :rtype: list[Blob]
        """
        blobs = self._base.get_blobs()
        blobs.extend(self._evidence.get_blobs())
        return blobs

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        jnode = {
            'uuid': self._uuid
        }

        jnode.update(self._base.serialize(tz_service))
        jnode.update(self._evidence.serialize(tz_service))

        return jnode


class Ticket:
    def __init__(self, name):
        """
        :type name: str
        """
        self.name = name


class Mask(list):
    def __init__(self, points):
        """
        :type points: list[(int, int)]
        """
        list.__init__(self, points)

    def normalize(self, image_w, image_h):
        """
        :type image_w: int
        :type image_h: int
        """
        raise NotImplementedError


class SceneArgs:
    def __init__(self, mask_args, mask_wh):
        """
        :type mask_args: Mask
        :type mask_wh: (int, int)
        """
        self.mask = mask_args
        self.mask_wh = mask_wh
